"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const type_1 = require("graphql/type");
const schemaInstrumentation_1 = require("./schemaInstrumentation");
const apollo_server_caching_1 = require("apollo-server-caching");
const dispatcher_1 = require("./dispatcher");
const schemaHash_1 = require("./schemaHash");
const graphql_1 = require("graphql");
const cachePolicy_1 = require("../cachePolicy");
async function pluginTestHarness({ pluginInstance, schema, logger, graphqlRequest, overallCachePolicy, executor, context = Object.create(null), }) {
    var _a;
    if (!schema) {
        schema = new type_1.GraphQLSchema({
            query: new type_1.GraphQLObjectType({
                name: 'RootQueryType',
                fields: {
                    hello: {
                        type: type_1.GraphQLString,
                        resolve() {
                            return 'hello world';
                        },
                    },
                },
            }),
        });
    }
    const schemaHash = schemaHash_1.generateSchemaHash(schema);
    let serverListener;
    if (typeof pluginInstance.serverWillStart === 'function') {
        const maybeServerListener = await pluginInstance.serverWillStart({
            logger: logger || console,
            schema,
            schemaHash,
            serverlessFramework: false,
            apollo: {
                key: 'some-key',
                graphRef: 'graph@current',
            },
        });
        if (maybeServerListener && maybeServerListener.serverWillStop) {
            serverListener = maybeServerListener;
        }
    }
    const requestContext = {
        logger: logger || console,
        schema,
        schemaHash: schemaHash_1.generateSchemaHash(schema),
        request: graphqlRequest,
        metrics: Object.create(null),
        source: graphqlRequest.query,
        cache: new apollo_server_caching_1.InMemoryLRUCache(),
        context,
        overallCachePolicy: cachePolicy_1.newCachePolicy(),
    };
    if (requestContext.source === undefined) {
        throw new Error('No source provided for test');
    }
    if (overallCachePolicy) {
        requestContext.overallCachePolicy.replace(overallCachePolicy);
    }
    if (typeof pluginInstance.requestDidStart !== 'function') {
        throw new Error('This test harness expects this to be defined.');
    }
    const listener = await pluginInstance.requestDidStart(requestContext);
    const dispatcher = new dispatcher_1.Dispatcher(listener ? [listener] : []);
    const executionListeners = [];
    await dispatcher.invokeHook('didResolveSource', requestContext);
    if (!requestContext.document) {
        await dispatcher.invokeDidStartHook('parsingDidStart', requestContext);
        try {
            requestContext.document = graphql_1.parse(requestContext.source, undefined);
        }
        catch (syntaxError) {
            const errorOrErrors = syntaxError;
            requestContext.errors = Array.isArray(errorOrErrors)
                ? errorOrErrors
                : [errorOrErrors];
            await dispatcher.invokeHook('didEncounterErrors', requestContext);
            await dispatcher.invokeHook('willSendResponse', requestContext);
            return requestContext;
        }
        const validationDidEnd = await dispatcher.invokeDidStartHook('validationDidStart', requestContext);
        const validationErrors = graphql_1.validate(requestContext.schema, requestContext.document);
        if (validationErrors.length !== 0) {
            requestContext.errors = validationErrors;
            validationDidEnd(validationErrors);
            await dispatcher.invokeHook('didEncounterErrors', requestContext);
            await dispatcher.invokeHook('willSendResponse', requestContext);
            return requestContext;
        }
        else {
            validationDidEnd();
        }
    }
    const operation = graphql_1.getOperationAST(requestContext.document, requestContext.request.operationName);
    requestContext.operation = operation || undefined;
    requestContext.operationName =
        (operation && operation.name && operation.name.value) || null;
    await dispatcher.invokeHook('didResolveOperation', requestContext);
    (await dispatcher.invokeHook('executionDidStart', requestContext)).forEach((executionListener) => {
        if (executionListener) {
            executionListeners.push(executionListener);
        }
    });
    executionListeners.reverse();
    const executionDispatcher = new dispatcher_1.Dispatcher(executionListeners);
    const invokeWillResolveField = (...args) => executionDispatcher.invokeSyncDidStartHook('willResolveField', ...args);
    Object.defineProperty(requestContext.context, schemaInstrumentation_1.symbolExecutionDispatcherWillResolveField, { value: invokeWillResolveField });
    schemaInstrumentation_1.enablePluginsForSchemaResolvers(schema);
    try {
        requestContext.response = await executor(requestContext);
        await executionDispatcher.invokeHook('executionDidEnd');
    }
    catch (executionErr) {
        await executionDispatcher.invokeHook('executionDidEnd', executionErr);
    }
    await dispatcher.invokeHook('willSendResponse', requestContext);
    await ((_a = serverListener === null || serverListener === void 0 ? void 0 : serverListener.serverWillStop) === null || _a === void 0 ? void 0 : _a.call(serverListener));
    return requestContext;
}
exports.default = pluginTestHarness;
//# sourceMappingURL=pluginTestHarness.js.map